PATH_OUTPUT_ROTGRIDS = 'output/rot_grids/'
PATH_OUTPUT_PT = 'output/pt_files/'
PATH_INPUT_BASEGRO = 'input/'
PATH_OUTPUT_PLOTS = 'output/figures/'
PATH_OUTPUT_ANIS = 'output/animations/'
PATH_OUTPUT_STAT = 'output/statistics_files/'
PATH_OUTPUT_TRANSGRIDS = 'output/trans_grids/'
